/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2 of
* the License, or (at your option) any later version.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU Lesser General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @package    RobotAPI::OpenHand
* @author     Markus Przybylski
* @date       2014 Markus Przybylski

* @copyright  http://www.gnu.org/licenses/gpl.txt
*             GNU General Public License
*/

#include "OpenHand.h"
//#include "Armar3GraspContext.h"   //MP: Maybe necessary again later?
#include "../../core/RobotStatechartContext.h"

#include <Core/observers/variant/ChannelRef.h>
#include <Core/observers/variant/SingleTypeVariantList.h>
#include <Core/robotstate/remote/ArmarPose.h>
#include <VirtualRobot/IK/DifferentialIK.h>

#include <Core/interface/units/KinematicUnitInterface.h>

#include <RobotAPI/core/RobotStatechartContext.h>

namespace armarx
{
    // ****************************************************************
    // Implementation of StatechartOpenHand
    // ****************************************************************
    void StatechartOpenHand::defineParameters()     //(OK)TODO: Anpassen!
    {
        //openHand settings:
        addToInput("jointAnglesOpenHand", VariantType::List(VariantType::Float), false);
        addToInput("timeoutOpenHandSuccess", VariantType::Float, false);
        addToInput("timeoutOpenHandDummyFailure", VariantType::Float, false);

        addToInput("robotNodeSetName", VariantType::String, false);
        addToLocal("jointNames", VariantType::List(VariantType::String));

        //addToLocal("jointVelocitiesDatafields", VariantType::List(VariantType::DatafieldRef));

    }

    void StatechartOpenHand::defineSubstates()     //(OK)TODO: Anpassen!
    {
        StatePtr stateOpenHand = addState<StateOpenHand>("stateOpenHand");
        StatePtr stateSuccess = addState<SuccessState>("stateSuccess");
        StatePtr stateFailure = addState<FailureState>("stateFailure");

        ParameterMappingPtr mapOpenHandInfo = ParameterMapping::createMapping()
                ->mapFromParent("jointAnglesOpenHand", "jointAnglesOpenHand")
                ->mapFromParent("timeoutOpenHandSuccess", "timeoutOpenHandSuccess")
                ->mapFromParent("timeoutOpenHandDummyFailure", "timeoutOpenHandDummyFailure")
                ->mapFromParent("jointNames", "jointNames");

        setInitState(stateOpenHand, mapOpenHandInfo);

        //transitions
        addTransition<EvOpenHandSuccessTimeout>(stateOpenHand, stateSuccess);
        addTransition<EvOpenHandDummyFailureTimeout>(stateOpenHand, stateFailure);  //Do we need to include a failure state?

    }

    void StatechartOpenHand::onEnter()     //TODO: Anpassen! ???
    {
        ARMARX_LOG << eINFO << "Entering StatechartOpenHand";

        RobotStatechartContext* context = getContext<RobotStatechartContext>();
        //setLocal("jointVelocityChannel", context->getChannelRef(context->getKinematicUnitObserverName(),"jointvelocities"));
        ChannelRefPtr tempChannelRef = context->getChannelRef(context->getKinematicUnitObserverName(),"jointvelocities");

        VirtualRobot::RobotNodeSetPtr nodeSet = context->remoteRobot->getRobotNodeSet(getInput<std::string>("robotNodeSetName"));    //woher robotNodeSetName holen? (Argument für getRobotNodeSet())

        SingleTypeVariantList jointNames(VariantType::String);
        SingleTypeVariantList dataFields(VariantType::DatafieldRef);


        for (size_t i=0; i<nodeSet->getSize(); i++)
        {
            jointNames.addVariant(nodeSet->getNode(i)->getName());      //nodes = joints
            dataFields.addVariant(context->getDatafieldRef(tempChannelRef, nodeSet->getNode(i)->getName()));
        }

        setLocal("jointNames", jointNames);
        setLocal("jointVelocitiesDatafields", dataFields);

    }

    void StatechartOpenHand::onExit()
    {

        ARMARX_LOG << eINFO << "Exiting StatechartOpenHand...";

    }

    // ****************************************************************
    // Implementation of StateOpenHand
    // ****************************************************************

    void StateOpenHand::defineParameters()      //(OK)TODO: Anpassen!
    {
        addToInput("jointAnglesOpenHand", VariantType::List(VariantType::Float), false);
        addToInput("timeoutOpenHandSuccess", VariantType::Float, false);

        addToInput("jointNames", VariantType::List(VariantType::String), false);
        //addToInput("useTorquesForGrasping", VariantType::Int, false);
    }

    void StateOpenHand::onEnter()      //(OK)TODO: Anpassen!
    {
        ARMARX_LOG << eIMPORTANT << "Entering StateOpenHand";

        RobotStatechartContext* rsContext = getContext<RobotStatechartContext>();
        SingleTypeVariantListPtr jointNames = getInput<SingleTypeVariantList>("jointNames");
        SingleTypeVariantListPtr jointAnglesOpenHandList = getInput<SingleTypeVariantList>("jointAnglesOpenHand");
        NameValueMap jointNamesAndValues;

        if (jointNames->getSize() == jointAnglesOpenHandList->getSize())
        {
            for (int j=0; j<jointNames->getSize(); j++)
            {
                jointNamesAndValues[jointNames->getVariant(j)->getString()]=jointAnglesOpenHandList->getVariant(j)->getFloat();
            }
        }
        else
            throw LocalException("StateOpenHand: List lengths do not match!");

        rsContext->kinematicUnitPrx->setJointAngles(jointNamesAndValues);

        ARMARX_LOG << eINFO << "Installing openHand timeout condition";
        float timeoutOpenHandSuccess = getInput<float>("timeoutOpenHandSuccess");

        condOpenHandSuccessTimeout = setTimeoutEvent(timeoutOpenHandSuccess, createEvent<EvOpenHandSuccessTimeout>());
        condOpenHandDummyFailureTimeout = setTimeoutEvent(timeoutOpenHandSuccess, createEvent<EvOpenHandDummyFailureTimeout>());

        ARMARX_LOG << eINFO << "StateOpenHand: Done onEnter()";
    }

    void StateOpenHand::onExit()      //(OK)TODO: Anpassen!
    {
        removeTimeoutEvent(condOpenHandSuccessTimeout);
        removeTimeoutEvent(condOpenHandDummyFailureTimeout);
        ARMARX_LOG << eIMPORTANT << "StateOpenHand: Done onExit()";
    }



}

