#include "HapticObserver.h"

#include <Core/observers/checks/ConditionCheckUpdated.h>
#include <Core/observers/checks/ConditionCheckEquals.h>
#include <Core/observers/checks/ConditionCheckInRange.h>
#include <Core/observers/checks/ConditionCheckLarger.h>
#include <Core/observers/checks/ConditionCheckSmaller.h>
#include <Core/robotstate/remote/checks/ConditionCheckMagnitudeChecks.h>

#include <Core/robotstate/remote/RobotStateObjectFactories.h>
#include <Eigen/Dense>

using namespace armarx;

HapticObserver::HapticObserver()
{
}

void HapticObserver::setTopicName(std::string topicName)
{
    this->topicName = topicName;
}

void HapticObserver::onInitObserver()
{
    if(topicName.empty())
        usingTopic(getProperty<std::string>("HapticTopicName").getValue());
    else
        usingTopic(topicName);

    // register all checks
    offerConditionCheck("updated", new ConditionCheckUpdated());
    offerConditionCheck("larger", new ConditionCheckLarger());
    offerConditionCheck("equals", new ConditionCheckEquals());
    offerConditionCheck("smaller", new ConditionCheckSmaller());

}

void HapticObserver::onConnectObserver()
{
}

void HapticObserver::reportSensorValues(const std::string &type, const MatrixFloatBasePtr &values, const Ice::Current &)
{
    ScopedLock lock(dataMutex);
    MatrixFloatPtr matrix = MatrixFloatPtr::dynamicCast(values);
    Eigen::MatrixXf eigenMatrix = matrix->toEigen();
    float max = eigenMatrix.maxCoeff();
    float mean = eigenMatrix.mean();
    if(!existsChannel(type))
    {
        offerChannel(type, "Force and Torque vectors on specific parts of the robot.");
        offerDataFieldWithDefault(type, "matrix", matrix, "Raw tactile matrix data");
        offerDataFieldWithDefault(type, "max", Variant(max), "Maximum value");
        offerDataFieldWithDefault(type, "mean", Variant(mean), "Mean value");
    }
    else
    {
        setDataField(type, "matrix", matrix);
        setDataField(type, "max", Variant(max));
        setDataField(type, "mean", Variant(mean));
    }

    updateChannel(type);
}

PropertyDefinitionsPtr HapticObserver::createPropertyDefinitions()
{
    return PropertyDefinitionsPtr(new HapticObserverPropertyDefinitions(getConfigIdentifier()));
}
